package de.hft.stuttgart.citydoctor2.connect.edge;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

public class CppHealResult extends CppReferenceHandlingBase {

	private native void delete(long ptr);

	private native int[] getChangedOrAddedVertexIds(long ptr);

	private native double[] getChangedOrAddedVertexValues(long ptr);

	private native int[] getChangedOrAddedPolygonIds(long ptr);

	private native int[] getChangedOrAddedPolygon(long ptr, int id);
	
	

	public CppHealResult(long ptr) {
		ptrToCppObject = ptr;
	}

	@Override
	public void disposeCppObject() {
		if (ptrToCppObject != 0) {
			delete(ptrToCppObject);
			ptrToCppObject = 0;
		}
	}

	public void mergeIntoGeometry(Geometry geom) {
		Map<Integer, Vertex> vertexMap = updateVertices(geom);
		
		int[] changedPolygons = getChangedOrAddedPolygonIds(ptrToCppObject);
		List<Polygon> deletedPolygons = new ArrayList<>();
		for (int i = 0; i < changedPolygons.length; i++) {
			int polyId = changedPolygons[i];
			int[] vertices = getChangedOrAddedPolygon(ptrToCppObject, polyId);
			if (vertices == null || vertices.length == 0) {
				if (polyId >= geom.getPolygons().size()) {
					throw new IllegalStateException("Cannot delete a polygon which does not exist");
				}
				Polygon p = geom.getPolygons().get(polyId);
				deletedPolygons.add(p);
			} else {
				Polygon p;
				if (polyId < geom.getPolygons().size()) {
					p = geom.getPolygons().get(polyId);
					// empty the polygon
					p.removeRings();
				} else {
					p = new ConcretePolygon();
					geom.addPolygon(p);
				}
				createPolygon(geom, vertexMap, vertices, p);
			}
		}
		for (Polygon p : deletedPolygons) {
			geom.getPolygons().remove(p);
		}
	}

	private void createPolygon(Geometry geom, Map<Integer, Vertex> vertexMap, int[] vertices, Polygon p) {
		LinearRing lr = null;
		for (int j = 0; j < vertices.length; j++) {
			if (vertices[j] == -1) {
				if (p.getExteriorRing() != null) {
					throw new IllegalStateException("Cannot create a polygon with more than 1 exterior rings");
				}
				lr = new LinearRing(LinearRingType.EXTERIOR);
				p.setExteriorRing(lr);
			} else if (vertices[j] == -2) {
				lr = new LinearRing(LinearRingType.INTERIOR);
				p.addInteriorRing(lr);
			} else {
				Objects.requireNonNull(lr, "No ring type specified, before supplying vertices");
				int index = vertices[j];
				Vertex v;
				if (index < geom.getVertices().size()) {
					v = geom.getVertices().get(index);
				} else {
					v = vertexMap.get(index);
				}
				Objects.requireNonNull(v, "Polygon references vertex, which does not exist");
				lr.addVertex(v);
			}
		}
		// close the rings
		p.getExteriorRing().addVertex(p.getExteriorRing().getVertices().get(0));
		for (LinearRing inner : p.getInnerRings()) {
			inner.addVertex(inner.getVertices().get(0));
		}
	}

	private Map<Integer, Vertex> updateVertices(Geometry geom) {
		int[] changedVertices = getChangedOrAddedVertexIds(ptrToCppObject);
		double[] changedVertexValues = getChangedOrAddedVertexValues(ptrToCppObject);
		Map<Integer, Vertex> vertexMap = new HashMap<>();
		for (int i = 0; i < changedVertices.length; i++) {
			int id = changedVertices[i];
			double x = changedVertexValues[i * 3 + 0];
			double y = changedVertexValues[i * 3 + 1];
			double z = changedVertexValues[i * 3 + 2];
			if (id >= geom.getVertices().size()) {
				// new vertex
				Vertex newV = new Vertex(x, y, z);
				vertexMap.put(id, newV);
				System.out.println("New Vertex: " + newV);
			} else {
				// change vertex
				Vertex changedV = geom.getVertices().get(id);
				System.out.print("Changed Vertex from " + changedV + " to ");
				changedV.setX(x);
				changedV.setY(y);
				changedV.setZ(z);
				System.out.println(changedV);
			}
		}
		return vertexMap;
	}

	public void print() {
		int[] changedVertices = getChangedOrAddedVertexIds(ptrToCppObject);
		System.out.println("Changed or added vertex ids: " + Arrays.toString(changedVertices));
		double[] changedVertexValues = getChangedOrAddedVertexValues(ptrToCppObject);
		if (changedVertices.length * 3 != changedVertexValues.length) {
			throw new IllegalStateException("Number of indices and number of values do not match");
		}
		for (int i = 0; i < changedVertices.length; i++) {
			System.out.println(String.format("Vertex %d = [%f, %f, %f]", changedVertices[i],
					changedVertexValues[i * 3 + 0], changedVertexValues[i * 3 + 1], changedVertexValues[i * 3 + 2]));
		}
		System.out.println();
		int[] changedPolygons = getChangedOrAddedPolygonIds(ptrToCppObject);
		System.out.println("Changed or added polygon ids: " + Arrays.toString(changedPolygons));
		for (int i = 0; i < changedPolygons.length; i++) {
			int polyId = changedPolygons[i];
			System.out.println("Polygon " + polyId + ":");
			int[] vertices = getChangedOrAddedPolygon(ptrToCppObject, polyId);
			if (vertices == null || vertices.length == 0) {
				System.out.println("deleted");
			} else {
				for (int j = 0; j < vertices.length; j++) {
					if (vertices[j] == -1) {
						System.out.println("Exterior ring:");
					} else if (vertices[j] == -2) {
						System.out.println("Interior ring:");
					} else {
						System.out.println(vertices[j]);
					}
				}
			}
		}
	}

	public boolean isEmpty() {
		int[] changedVertices = getChangedOrAddedVertexIds(ptrToCppObject);
		int[] changedPolygons = getChangedOrAddedPolygonIds(ptrToCppObject);
		if(changedVertices.length == 0 && changedPolygons.length == 0)
			return true;
	
		return false;
	}
	
	
}
